#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import MutableMapping, Callable, Optional, cast
from weakref import WeakKeyDictionary

from sqlalchemy.orm import Relationship, InstrumentedAttribute

relation_to_descend: MutableMapping[Relationship, Callable[[Optional[str]], 'EntityProxy']] = WeakKeyDictionary()
relation_to_ascend: MutableMapping[Relationship, Callable[[str], 'EntityProxy']] = WeakKeyDictionary()


def descend(attribute: InstrumentedAttribute, alias: Optional[str] = None) -> 'EntityProxy':
    relation: Relationship = cast(Relationship, attribute.property)
    if descend_ := relation_to_descend.get(relation):
        return descend_(alias)
    else:
        raise KeyError(f"{relation} not found in relation_to_descend mapping")


def ascend(attribute: InstrumentedAttribute, alias: str) -> 'EntityProxy':
    relation: Relationship = cast(Relationship, attribute.property)
    if ascend_ := relation_to_ascend.get(relation):
        return ascend_(alias)
    else:
        raise KeyError(f"{relation} not found in relation_to_descend mapping")



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import copy
import dataclasses
from dataclasses import dataclass, Field
from logging import getLogger
from typing import Optional, Callable

import sqlalchemy
import sqlalchemy.orm
from init_helpers.dict_to_dataclass import get_dataclass_field_name_to_field, NoValue

from entity_read.entity import Entity
from entity_read.entity_field import get_sqlalchemy_metadata
from .utils import is_list_relation, has_default

logger = getLogger(__name__)


@dataclass(unsafe_hash=True, repr=False)
class _InsertViewMixin:
    @classmethod
    def _produce_field_name_to_field(cls, related_entity_type: type[Entity],
                                     relation_remote_column: sqlalchemy.Column = None) -> dict[str, Field | Callable]:
        # sorry for this: fast way to break circular dependency
        from .relation_insert import RelationInsertView
        field_name_to_field = copy.copy(get_dataclass_field_name_to_field(related_entity_type))
        result = {}
        for field_name, field in field_name_to_field.items():
            field_copy = copy.copy(field)
            sql_alchemy_property = get_sqlalchemy_metadata(field)

            if isinstance(sql_alchemy_property, sqlalchemy.orm.RelationshipProperty):
                relation: sqlalchemy.orm.RelationshipProperty = sql_alchemy_property
                field_copy.default = NoValue
                field_copy.default_factory = dataclasses.MISSING
                related_entity_view_type = RelationInsertView[relation]
                is_list = is_list_relation(relation)
                field_copy.type = list[related_entity_view_type] if is_list else related_entity_view_type
            elif isinstance(sql_alchemy_property, sqlalchemy.Column):
                column: sqlalchemy.Column = sql_alchemy_property
                if relation_remote_column is not None and column == relation_remote_column:
                    field_copy.init = False
                    field_copy.default = NoValue
                if field.default in (dataclasses.MISSING, NoValue) and has_default(column):
                    field_copy.default = NoValue
                    # used to consider null value as missing, if null is prohibited by structure; example: pk
                    if not column.nullable:
                        field_copy.type = Optional[field_copy.type]
            else:
                logger.warning(f"ignored attribute: %s, cos %s has unexpected type", field, sql_alchemy_property)

            result[field_name] = field_copy

        result['entity_type'] = related_entity_type
        return result



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
from dataclasses import dataclass, field
from typing import ClassVar, Any

from ..atoms import Function
from ..atoms.abstract import Selectable


@dataclass(kw_only=True, frozen=True, slots=True)
class BooleanFunction(Function, abc.ABC):
    type_: str = field(default=bool, init=False)


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionStringContains(BooleanFunction):
    args: tuple[Selectable[Any], Selectable[str]]
    code: ClassVar[str] = 'string_contains'
    ignore_case: bool = False

    def _to_sql(self, args_sql: list[str]) -> str:
        return f"{args_sql[0]} {'i' if self.ignore_case else ''}like ('%' || {args_sql[1]} || '%')"


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionIsNull(BooleanFunction):
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'is_null'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} IS NULL'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionIsNotNull(BooleanFunction):
    args: tuple[Selectable[Any]]
    code: ClassVar[str] = 'is_not_null'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} IS NOT NULL'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionLessThan(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[int | float]]
    code: ClassVar[str] = 'lt'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} < {args_sql[1]}'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionLessEqual(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[int | float]]
    code: ClassVar[str] = 'le'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} <= {args_sql[1]}'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionEqual(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[int | float]]
    code: ClassVar[str] = 'eq'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} = {args_sql[1]}'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionNonEqual(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[int | float]]
    code: ClassVar[str] = 'ne'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} != {args_sql[1]}'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionGreaterEqual(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[int | float]]
    code: ClassVar[str] = 'ge'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} >= {args_sql[1]}'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionGreater(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[int | float]]
    code: ClassVar[str] = 'gt'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} > {args_sql[1]}'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionIn(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[list]]
    code: ClassVar[str] = 'in'

    def _to_sql(self, args_sql: list[str]) -> str:
        if args_sql[1] != "'{}'":
            return f'{args_sql[0]} = ANY ({args_sql[1]})'
        return f'FALSE'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionNotIn(BooleanFunction):
    args: tuple[Selectable[int | float], Selectable[list]]
    code: ClassVar[str] = 'not_in'

    def _to_sql(self, args_sql: list[str]) -> str:
        if args_sql[1] != "'{}'":
            return f'{args_sql[0]} != ALL ({args_sql[1]})'
        return f'TRUE'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionOr(BooleanFunction):
    args: tuple[Selectable[bool], ...]
    code: ClassVar[str] = 'or'

    def _to_sql(self, args_sql: list[str]) -> str:
        if not args_sql:
            return 'TRUE'
        if len(args_sql) > 1:
            return f"({' OR '.join(args_sql)})"
        return args_sql[0]


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionAnd(BooleanFunction):
    args: tuple[Selectable[bool], ...]
    code: ClassVar[str] = 'and'

    def _to_sql(self, args_sql: list[str]) -> str:
        if not args_sql:
            return 'TRUE'
        if len(args_sql) > 1:
            return f"({' AND '.join(args_sql)})"
        return args_sql[0]


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionNot(BooleanFunction):
    args: tuple[Selectable[bool]]
    code: ClassVar[str] = 'not'

    def _to_sql(self, args_sql: list[str]) -> str:
        if not args_sql:
            return 'FALSE'
        if len(args_sql) > 1:
            raise ValueError(f'{type(self).__class__} expected one arg, got: {len(args_sql)}')
        return f'NOT ({args_sql[0]})'


@dataclass(kw_only=True, frozen=True, slots=True)
class FunctionBetween(BooleanFunction):
    args: tuple[Selectable[int | float | str], Selectable[int | float | str], Selectable[int | float | str]]
    code: ClassVar[str] = 'between'

    def _to_sql(self, args_sql: list[str]) -> str:
        return f'{args_sql[0]} BETWEEN {args_sql[1]} AND {args_sql[2]}'
